@@ -60,7 +60,8 @@ class MessageMeta(type):
         # shorthand for a nested message and a repeated field of that message.
         # Decompose each map into its constituent form.
         # https://developers.google.com/protocol-buffers/docs/proto3#maps
-        for key, field in copy.copy(attrs).items():
+        map_fields = {}
+        for key, field in attrs.items():
             if not isinstance(field, MapField):
                 continue
 
@@ -92,13 +93,16 @@ class MessageMeta(type):
             entry_attrs["value"] = Field(
                 field.proto_type, number=2, enum=field.enum, message=field.message,
             )
-            attrs[msg_name] = MessageMeta(msg_name, (Message,), entry_attrs)
+            map_fields[msg_name] = MessageMeta(msg_name, (Message,), entry_attrs)
 
             # Create the repeated field for the entry message.
-            attrs[key] = RepeatedField(
-                ProtoType.MESSAGE, number=field.number, message=attrs[msg_name],
+            map_fields[key] = RepeatedField(
+                ProtoType.MESSAGE, number=field.number, message=map_fields[msg_name],
             )
 
+        # Add the new entries to the attrs
+        attrs.update(map_fields)
+
         # Okay, now we deal with all the rest of the fields.
         # Iterate over all the attributes and separate the fields into
         # their own sequence.
@@ -254,7 +258,7 @@ class MessageMeta(type):
 
         # Generate the descriptor for the file if it is ready.
         if file_info.ready(new_class=cls):
-            file_info.generate_file_pb()
+            file_info.generate_file_pb(new_class=cls, fallback_salt=full_name)
 
         # Done; return the class.
         return cls
@@ -316,7 +320,11 @@ class MessageMeta(type):
             ~.Message: An instance of the message class against which this
             method was called.
         """
-        return cls(cls.pb().FromString(payload))
+        # Usually we don't wrap the original proto and are force to make a copy
+        # to prevent modifying user data.
+        # In this case it's perfectly reasonable to wrap the proto becasue it's
+        # never user visible, and it gives a huge performance boost.
+        return cls(cls.pb().FromString(payload), __wrap_original=True)
 
     def to_json(cls, instance) -> str:
         """Given a message instance, serialize it to json
@@ -341,7 +349,9 @@ class MessageMeta(type):
             ~.Message: An instance of the message class against which this
             method was called.
         """
-        return cls(Parse(payload, cls()._pb))
+        instance = cls()
+        Parse(payload, instance._pb)
+        return instance
 
 
 class Message(metaclass=MessageMeta):
@@ -360,20 +370,9 @@ class Message(metaclass=MessageMeta):
         #   * An instance of the underlying protobuf descriptor class.
         #   * A dict
         #   * Nothing (keyword arguments only).
-        #
-        # Sanity check: Did we get something not on that list? Error if so.
-        if mapping and not isinstance(
-            mapping, (collections.abc.Mapping, type(self), self._meta.pb)
-        ):
-            raise TypeError(
-                "Invalid constructor input for %s: %r"
-                % (self.__class__.__name__, mapping,)
-            )
 
         # Handle the first two cases: they both involve keeping
         # a copy of the underlying protobuf descriptor instance.
-        if isinstance(mapping, type(self)):
-            mapping = mapping._pb
         if isinstance(mapping, self._meta.pb):
             # Make a copy of the mapping.
             # This is a constructor for a new object, so users will assume
@@ -390,27 +389,38 @@ class Message(metaclass=MessageMeta):
             if kwargs:
                 self._pb.MergeFrom(self._meta.pb(**kwargs))
             return
+        if isinstance(mapping, type(self)):
+            # Performance hack to streamline construction from vanilla protos.
+            self.__init__(mapping=mapping._pb, **kwargs)
+            return
 
         # Handle the remaining case by converging the mapping and kwargs
         # dictionaries (with kwargs winning), and saving a descriptor
         # based on that.
+
         if mapping is None:
             mapping = {}
+        elif not isinstance(mapping, collections.abc.Mapping):
+            # Sanity check: Did we get something not a map? Error if so.
+            raise TypeError(
+                "Invalid constructor input for %s: %r"
+                % (self.__class__.__name__, mapping,)
+            )
         mapping.update(kwargs)
 
+        # Avoid copying the mapping unnecessarily
+        params = {}
         # Update the mapping to address any values that need to be
         # coerced.
         marshal = self._meta.marshal
-        for key, value in copy.copy(mapping).items():
+        for key, value in mapping.items():
             pb_type = self._meta.fields[key].pb_type
             pb_value = marshal.to_proto(pb_type, value)
-            if pb_value is None:
-                mapping.pop(key)
-            else:
-                mapping[key] = pb_value
+            if pb_value is not None:
+                params[key] = pb_value
 
         # Create the internal protocol buffer.
-        self._pb = self._meta.pb(**mapping)
+        self._pb = self._meta.pb(**params)
 
     def __bool__(self):
         """Return True if any field is truthy, False otherwise."""
@@ -498,10 +508,13 @@ class Message(metaclass=MessageMeta):
             their Python equivalents. See the ``marshal`` module for
             more details.
         """
-        pb_type = self._meta.fields[key].pb_type
-        pb_value = getattr(self._pb, key)
-        marshal = self._meta.marshal
-        return marshal.to_python(pb_type, pb_value, absent=key not in self)
+        try:
+            pb_type = self._meta.fields[key].pb_type
+            pb_value = getattr(self._pb, key)
+            marshal = self._meta.marshal
+            return marshal.to_python(pb_type, pb_value, absent=key not in self)
+        except KeyError as ex:
+            raise AttributeError(str(ex))
 
     def __ne__(self, other):
         """Return True if the messages are unequal, False otherwise."""
@@ -516,7 +529,7 @@ class Message(metaclass=MessageMeta):
         For well-known protocol buffer types which are marshalled, either
         the protocol buffer object or the Python equivalent is accepted.
         """
-        if key.startswith("_"):
+        if key[0] == "_":
             return super().__setattr__(key, value)
         marshal = self._meta.marshal
         pb_type = self._meta.fields[key].pb_type
